/* JavaGame is a game engine used to teach computer programming.
 * There is a corresponding JavaScript version to use in a web page.
 *
 * JavaGame copyright 2020 Itty Bitty Computers and released at this time
 * to the public as open source. There are no warranties of any kind.
 */

/**
 * GameEvent is the base class to handle GameEngine events. -- 2020 July 24
 *
 * Use GameEngine's GameMaker to Build a subclass of GameEvent
 * in order to use the GameEngine in a game.
 */
public class GameEvent { // Game event base class

 /**
  * These constants are used internally in this base class. They are copied
  * from class JavaGame. If you need them, you can get them from there.
  */
  private final static boolean Logging = JavaGame. Logging;
  private final static char Arro_LEFT = JavaGame. Arro_LEFT,
      Arro_RIGHT = JavaGame. Arro_RIGHT,
      Arro_UP = JavaGame. Arro_UP,
      Arro_DOWN = JavaGame. Arro_DOWN,
      Esc_Key = JavaGame. Esc_Key;
  private final static int Clik2time = JavaGame. Clik2time;
  private static int MaxLogWgts = JavaGame. Max_Log,
      FustKeyLog = JavaGame. Fust_Log,
      MaxLogAnim = FustKeyLog>>1, SeenLog = 0;

 /**
  * EvsDlogOp is a local copy of DlogIsOpen, to make dialog events easy
  * to detect.
  *
  * ClickLocn is set here by FocusEv and MouseDn events, and cleared (to -1)
  * by BlurEv and MouseUp. It holds the vertical,horizontal offsets
  * (16 bits each) into the FocusWgt where it was clicked.
  *
  * CluckTime is captured from the game timer (see JavaGame.GetMilliSecs)
  * by the ClickEvt event and is used to recognize double-clicks (see
  * JavaGame.Clik2time) and is cleared to 0 when a double-click is recognized.
  *
  * DrugWgtBase is set by the MouseDn event for any draggable widget, and is
  * the starting position of that widget within its parent (used to return
  * that widget to its original location if the drag is cancelled).
  *
  * DrugWgtOfx is set by the MouseDn event for any draggable widget, and is
  * the ClickLocn for a widget being dragged. Dragging a widget consists in
  * continuously adjusting the position of that widget within its parent
  * so that the difference between the widget position and the widget-relative
  * mouse location remains constant and equal to DrugWgtOfx.
  *
  * DrugWgtGrid can be set by user code and if it is a non-zero power of two,
  * any widget being dragged will snap to the nearest grid point within its
  * parent. [This is untested]
  *
  * If non-zero, vDragLims and hDragLims are the vertical and horizontal limits
  * beyond which the user cannot drag a widget within it parent. The minimum
  * position is in the high 16 bits, and the maximum position (which cannot be
  * negative) is in the low 16 bits. If the specified minimum exceeds the
  * maximum, it will dominate. [The limits are untested at this time, but
  * should be valid when the scrollbars are implemented]
  */
  int EvsDlogOp, ClickLocn, CluckTime, // these are visible in subclass
      DrugWgtBase, DrugWgtOfx, DrugWgtGrid, vDragLims, hDragLims;

  /**
  * ClickWgt is set before the FocusEv and MouseDn events, to whatever widget
  * the user clicked, if that widget can accept MouseDn or ClickEvt events.
  *
  * RollyWgt is set for a widget that accepts rollover events, while the mouse
  * is currently over that widget. [This is untested]
  *
  * WgtBeingDrug is set by the (default) StartDrag event handler, and cleared
  * when the drag is terminated.
  *
  * FocusWgt is set by the (default) FocusEv event handler, and cleared by
  * the (default) BlurEv handler. It is normally used to direct keystrokes
  * to whatever widget wants them.
  *
  * DefaultFocus is set by user game code using the DefaultKeyFocus method,
  * and defines a default widget to receive keystrokes when there is no
  * FocusWgt (which takes precedence).
  *
  * Additional information is in the JavaDoc for GetRefWgt/SetRefWgt methods.
  */
  GameWgt ClickWgt, RollyWgt, Wgt2BeDrug, WgtBeingDrug, FocusWgt, DefaultFocus;

  /**
  * theGame is a local reference to the host JavaGame, set at startup.
  */
  static JavaGame theGame;

  // Handy (private) methods..

  private int GetMilliSecs() {return JavaGame.GetMilliSecs();}

  private GameWgt GetIxWgt(int ix) {return JavaGame.GetIxWgt(ix);}

/**
  * A safe (no exceptions) non-OOPS way to extract a character from a string.
  *
  * @param  here  The (0-base) character position
  * @param  aStr  The string from which to extract that character
  *
  * @return       The character extracted, or '\0' if out of bounds
  */
  private char CharAt(int here, String aStr)  { // Java version..
    if (here<0) return '\0';
    if (aStr == null) return '\0';
    try {
      if (here >= aStr.length()) return '\0';
      return aStr.charAt(here);}
    catch (Exception ex)  {return '\0';}} //~CharAt

 /**
  * A safe (no exceptions) non-OOPS way to get the length of a string.
  *
  * @param  aStr  The string to get the length of
  *
  * @return       Its length, or 0 if null
  */
  private int StrLength(String aStr)  { // Java version..
    if (aStr == null) return 0;
    return aStr.length();} //~StrLength

 /**
  * Strongly-typed (not overloaded) string selector.
  * This is part of a collection of tools to build debug print lines.
  *
  * @param  whom  The selector
  * @param  tru   Returned if whom=true
  * @param  fls   Returned if whom=false
  *
  * @return       The selected string, tru if whom=true, else fls
  */
  private String IffyStr(boolean whom, String tru, String fls)  {
    if (whom) return tru;
    return fls;} //~IffyStr

 /**
  * A safe (no exceptions) way to get the rest of a string.
  * Text past the end is considered empty.
  *
  * @param  here  The (0-base) position to start
  * @param  aStr  The string from which to extract that substring
  *
  * @return       The rest of the string, after here
  */
  private String RestOf(int here, String aStr)  {
    int lxx;
    try {
    if (here<0) return "";
    lxx = StrLength(aStr);
    if (lxx==0) return "";
    if (here >= lxx) return "";
    return aStr.substring(here);
    } catch (Exception ex)  {}
    return "";  } //~RestOf

  /**
  * A safe (no exceptions) way to extract a substring.
  * Text past the end is empty.
  *
  * @param  here  The (0-base) position to start
  * @param  lxx   The desired length (the result could be less)
  * @param  aStr  The string from which to extract that substring
  *
  * @return       The substring
  */
  private String Substring(int here, int lxx, String aStr)  {
    int nx = StrLength(aStr);
    try {
    if (here<0) return "";
    if (lxx <= 0) return "";
    if (nx==0) return "";
    if (here >= nx) return "";
    lxx = here+lxx;
    if (lxx >= nx) return RestOf(here,aStr);
    return aStr.substring(here,lxx);
    } catch (Exception ex)  {}
    return "";  } //~Substring

  /**
  * Formats an integer as hexadecimal and returns it with a prefix and suffix.
  * This is useful when printing numbers that you want to see the bits of.
  *
  * @param  before The prefix
  * @param  whom   The number to be formatted
  * @param  after  The suffix
  *
  * @return       The combined string
  */
  private String SeeHex(String before, int whom, String after)  {
    int shft = 28, lxx = StrLength(before); // special-case URL encoding..
    if (lxx==1) if ((whom&-256)==0) if (StrLength(after)==0)
        if ((int)(CharAt(0,before)) == 37)  { // '%'
      if (whom>15)  {
        before = before + SeeHex("",whom>>4,"");
        whom = whom&15;} //~if
      else before = before + "0";} //~if
    if (whom<0)  {
      before = SeeHex(before,(whom>>28)&15,"");
      if (whom+0x00100000 >= 0)  { // remove leading 'F's..
        shft = 16;
        while ((1<<shft)+whom >= 0)  {
          shft = shft-4;
          if (shft==0) break;} //~while
        while (shft>0)  { // skip over (internal) leading zeros..
          if (((15<<shft)&whom) !=0) break;
          shft = shft-4;} //~while
        whom = whom&((16<<shft)-1);
        before = before + "-F";} //~if
      else whom = whom&0x0FFFFFFF;} //~if // (whom<0)
    if (whom>15) before = SeeHex(before,whom>>4,"");
    return before + Substring(whom&15,1,"0123456789ABCDEF")
        + after;} //~SeeHex

  /**
  * Extracts and sign-extends the second of a pair of half-integers.
  *
  * @param  here  The pair
  *
  * @return       The low half, sign-extended to a full int
  */
  private int SignExtend(int here)  { // low 16 bits signed
    return (here&0x7FFF)-(here&0x8000);} //~SignExtend

  /**
  * Formats a 32-bit integer as two 16-bit values and returns it with a prefix
  * and suffix. This is useful when printing numbers that pack two smaller
  * integers (like board coordinates) into a single number.
  *
  * @param  before The prefix
  * @param  whom   The number to be formatted
  * @param  after  The suffix
  *
  * @return       The combined string
  */
  private String LogVH(String before, int whom, String after)  {
    return before + (whom>>16) + "/" + SignExtend(whom) + after;} //~LogVH

/**
  * Adds two pairs of half-integers in a single int.
  *
  * @param  here  The first pair
  * @param  thar  The second pair
  *
  * @return       The pair of sums
  */
  private int AddPair(int here, int thar)  {
    return (here&-0x10000)+(thar&-0x10000)
        + ((here+thar)&0xFFFF);} //~AddPair

  /**
  * Returns in a single int the negatives of a pair of half-integers.
  *
  * @param  here  The pair
  *
  * @return       Its negative
  */
  private int PairNeg(int here)  {
    return ((-here)&0xFFFF)-(here&-0x10000);} //~PairNeg

/**
  * A safe (no exceptions) way to get the size of a (possibly null) array.
  *
  * @param  data  The array
  *
  * @return       Its length, or =0 if null
  */
  private int SafeArySz(int[] data)  {
    if (data == null) return 0;
    return data.length;} //~SafeArySz

 /**
  * A safe (no exceptions) way to get a (possibly null) array element.
  * Returns a specified default value if the array is null or the given
  * index is out of bounds.
  *
  * @param  ix      The index into the array
  * @param  data    The array
  * @param  defalt  A default value to return if can't access
  *
  * @return       Its length, or =0 if null
  */
  private int SafeAryElt(int ix, int[] data, int defalt)  {
    while (true)  {
      if (data == null) break;
      if (ix<0) break;
      if (ix >= data.length) break;
      defalt = data[ix];
      break;} //~while
    return defalt;} //~SafeAryElt

  /**
  * Formats an integer as hexadecimal if larger than 32K and returns it
  * with a prefix and suffix, for easier reading.
  *
  * @param  before The prefix
  * @param  whom   The number to be formatted
  * @param  after  The suffix
  *
  * @return       The result string
  */
  private String HexIfMore(String before, int whom, String after)  {
    if (SignExtend(whom)==whom) return before + whom + after;
    return SeeHex(before + "0x",whom,after); } //~HexIfMore

  /**
  * Formats a 16-bit integer as 8.8 fixed-point and returns it with a prefix.
  *
  * @param  before The prefix
  * @param  whom   The number to be formatted
  *
  * @return       The result string
  */
  private String FormFixt(String before, int whom)  {
    int frac = whom&255, zx = frac*10+6;   // assumes 8-bit fraction
    String info = "";                     // ..and signed 8+bit integer part
    boolean shorten = ((whom>25599)||(StrLength(before)==1));
    if (false) if ((whom&-0x8000) !=0) if (whom+0x10000 != (whom&0xFFFF))
      whom = SignExtend(whom); // caller did
    if (whom<0) return FormFixt(before + "-",-whom);
    before = before + (whom>>8);
    if (frac>0)  {
      before = before + ".";
      if (frac==128) info = "5";
      else if (frac==64) info = "25";
      else if (frac==192) info = "75";
      else if (frac==1)  {
        if (shorten) info = "01"; else
          info = "004";} //~if
      else if ((zx&255)<13) info = info + (zx>>8); // rounds to 1 digit
      else if (shorten)  { // round to 2 fractional digits..
        frac = (frac*100+127)>>8;
        if (frac<10) info = "0";
        else if (frac>99) frac = 99;
        info = info + frac;} //~if
      else if ((frac&31) !=0)  { // round to 3 fractional digits..
        frac = (frac*1000+127)>>8;
        if (frac<10) info = "00";
        else if (frac<100) info = "0";
        else if (frac>999) frac = 999;
        info = info + frac;} //~if
      else info = info + ((frac>>5)*125);} //~if
    return before + info;} //~FormFixt

  // GameEvent public methods..

 /**
  * Return the class name for this object.
  *
  * @return     The reference String for this object
  */
  public String toString() {return "(GameEvent)";} //~toString

  // Access methods..

  /**
  * OpCloEvDlog is called when a standard dialog (alert or input) is
  * opened or closed. Override this method (but be sure to call super)
  * if your subclass wants to be informed.
  *
  * @param  info    the RefNum of the opening dialog group, =0 if closing
  */
  public void OpCloEvDlog(int info) {
    if (Logging) System.out.println("  (OpCloEvDlg=" + info + ")" );
    EvsDlogOp = info;} //~OpCloEvDlog

  /**
  * GetClickLoc returns the vertical,horizontal offsets (16 bits each) into
  * the FocusWgt where it was clicked, or -1 after the mouse is released or
  * the widget has lost its focus.
  *
  * ClickLocn is set by FocusEv, and cleared (to -1) by BlurEv.
  */
  public int GetClickLoc() {return ClickLocn;} //~GetClickLoc

  /**
  * DefaultKeyFocus defines a widget to receive keystrokes when no widget
  * has focus. Pass it null to cancel the setting.
  *
  * @param  whom    The widget to have default focus
  */
  public void DefaultKeyFocus(GameWgt whom) {
    if (Logging) System.out.println("  (DeFoc=" + JavaGame.  SafeWgtNum(whom) + ")" );
    DefaultFocus = whom;} //~DefaultKeyFocus

  /**
  * GetRefWgt can be called at any time to determine which widget has focus
  * (is being clicked or dragged or typeable). Use SetRefWgt to set them.
  *
  * Four widget references are maintained by the GameEngine to support
  * these events:
  *
  * FocusWgt is generally set by FocusEv and cleared by BlurEv, but a widget
  * can refuse to give up focus (or BlurEv can refuse for it), in which case
  * FocusEv will fail. FocusEv can also refuse to grant focus, or the widget
  * can refuse to accept it, or it can be flagged as "no-focus" (see GameScro)
  * and the low-level mouse event handler won't even try. You should avoid
  * overriding the default GameEngine focus behavior unless you really need
  * it, because it's very easy to make the rules so complicated as to be
  * inconsistent, which will make your game hard to play.
  *
  * ClickWgt is generally set by mouse-down and cleared by mouse-up, and is
  * used to determine if a mouse click should be reported as a ClickEvt. It is
  * also used to decide when to start a drag. If you change the default
  * behavior, these operations may misbehave.
  *
  * RollyWgt is generally set for a widget that accepts rollover events, and
  * cleared when the mouse exits the rollover. If you change the default
  * behavior, widget rollover may misbehave.
  *
  * WgtBeingDrug is generally set by StartDrag and cleared by MouseUp or
  * CancelDrag, and is used to control dragging behavior. If you change the
  * default behavior, widget dragging may misbehave.
  *
  * @param  why   A number used to select which widget you want:
  *               -1=FocusWgt, 0=RollyWgt, 1=ClickWgt, 2=WgtBeingDrug
  *
  * @return       The selected widget, or null if none.
  */
  public GameWgt GetRefWgt(int why) { // -1:Foc, 0:Rollo, 1:Clik, 2:Drg
    int rx = 0, zx = 0;
    if (why+2==0) { // -2 gets the DefaultFocus wgt
      if (FocusWgt != null) {
        rx = FocusWgt.GetRefNum();
        zx = FocusWgt.GetType();
        if (Logging) System.out.println("** GetFocWg #" + rx + " +" + zx);
        if ((zx&512) !=0)
          return FocusWgt;} //~if
      if (Logging) {
        rx = 0;
        zx = 0;
        if (DefaultFocus != null) {
          rx = DefaultFocus.GetRefNum();
          zx = DefaultFocus.GetType();} //~if
        System.out.println("** GetDFocWg #" + rx + " +" + zx);} //~if
      return DefaultFocus;} //~if
    if (why<0) return FocusWgt;
    if (why==0) return RollyWgt;
    if (why==1) return ClickWgt;
    if (why==2) return WgtBeingDrug;
    if (why==3) return Wgt2BeDrug; // (use ClickWgt instead)
    return null;} //~GetRefWgt

  /**
  * SetRefWgt can be called at any time to set which widget has focus
  * (is being clicked or dragged or typeable). Use GetRefWgt to get the
  * current settings.
  *
  * Four widget references are maintained by the GameEngine to support
  * these events:
  *
  * FocusWgt is generally set by FocusEv and cleared by BlurEv, but a widget
  * can refuse to give up focus (or BlurEv can refuse for it), in which case
  * FocusEv will fail. FocusEv can also refuse to grant focus, or the widget
  * can refuse to accept it, or it can be flagged as "no-focus" (see GameScro)
  * and the low-level mouse event handler won't even try. You should avoid
  * overriding the default GameEngine focus behavior unless you really need
  * it, because it's very easy to make the rules so complicated as to be
  * inconsistent, which will make your game hard to play.
  *
  * ClickWgt is generally set by mouse-down and cleared by mouse-up, and is
  * used to determine if a mouse click should be reported as a ClickEvt. It is
  * also used to decide when to start a drag. If you change the default
  * behavior, these operations may misbehave.
  *
  * RollyWgt is generally set for a widget that accepts rollover events, and
  * cleared when the mouse exits the rollover. If you change the default
  * behavior, widget rollover may misbehave.
  *
  * WgtBeingDrug is generally set by StartDrag and cleared by MouseUp or
  * CancelDrag, and is used to control dragging behavior. If you change the
  * default behavior, widget dragging may misbehave.
  *
  * @param  why   A number used to select which widget you want to set:
  *               -1=FocusWgt, 0=RollyWgt, 1=ClickWgt, 2=WgtBeingDrug
  * @param  whom  The widget you want to set it to, or else null.
  */
  public void SetRefWgt(int why, GameWgt whom) { // -1:Fo, 0:Ro, 1:Cl, 2:Dr
    int refn = 0, doit = why&3;
    if (whom != null) refn = whom.GetRefNum();
    if (Logging) System.out.println("** SetWgt #" + refn + " = " + why);
    if (why<0) FocusWgt = whom; // see also Focus/Blur
    else if (doit==1) ClickWgt = whom;
    else if (doit==0) RollyWgt = whom;
    else if (doit==2) WgtBeingDrug = whom;
    else if (doit==3) Wgt2BeDrug = whom;} //~SetRefWgt

  /**
  * You can override GetValue/SetValue in your subclass if you need access
  * to something we didn't think of.
  */
  public int GetValue(int why) {return 0;} //~GetValue // o'ride to use
  public void SetValue(int why, int whom) {} //~SetValue // o'ride to use

  /**
  * GameList returns a text descriptor of all the specified widgets in
  * a game. The code is normally generated when you click the "Build"
  * button in GameMaker. It is called once before StartUp, which builds
  * the widgets from this text descriptor list. You must call the super
  * to preserve the reference to the host JavaGame.
  *
  * @param  whom  A reference to the host JavaGame, which is kept in theGame
  *
  * @return       The generated text string
  */
  public String GameList(JavaGame whom) { //                   // (GameEv)
    String msg = "";
    boolean doit = false, prio = (theGame == null);
    if (Logging) {
      if (whom != null) msg = whom.toString();
      if (StrLength(msg)>0) {
        msg = "  GmEv.GmList: " + msg;
        doit = true;} //~if
      else msg = "  GmEv.GmList: (none)";
      System.out.println(msg + IffyStr(prio,IffyStr(doit," / new",
          " / (none)"),IffyStr(doit," / (2nd)"," / (got)")));} //~if
    if (prio) theGame = whom;
    return "";} //~GameList // called once to get your wgts list

  /**
  * After the initial widgets have been created, StartUp is called to do
  * any other initialization the client game may have.
  *
  * StartUp is normally overridden in the client subclass.
  *
  * If you gave Java-correct names to any widgets in GameMaker, those
  * variables are assigned values by the generated code here.
  */
  public void StartUp() {} //~StartUp // called once to set your game up

  /**
  * DoOften is called once each frame to do whatever game processing
  * needs to be done often. You should get in and get out, so that the
  * animation flows smoothly. Omit overriding this method if unneeded.
  */
  public void DoOften() {} //~DoOften // called once each frame (get in, get out)

  /**
  * GameEnding is called once when the game window is about to close,
  * so you can do whatever needs to be done at the end. Omit overriding
  * this method if unneeded. [GameEnding is not known to work]
  */
  public void GameEnding() {} //~GameEnding // called when the game is quitting

  /**
  * UserMacro is called for user-specified widget macros in the GameList.
  * See the JavaGame documentation for a complete explanation of macros.
  * [This mechanism is untested]
  *
  * @param  why   control number
  * @param  mID   macro ID code from source line
  * @param  whom  reference first to the parent group widget,
  *               then to each successor widget as created
  *
  * @return       the number of widgets to allocate, then info about each wgt
  */
  public int UserMacro(int why, int mID, GameWgt whom) {
    // called for user-spec macros, whom is init'ly container, then new wgt;
    // mID is `xxxx` from data line (use w/why<0 to distinguish)
    // why<0: rtn number n of subwgts to allocate (must be 0<n<256)+x*256
    //   where 0<=x<256 is used in subsequent calls to distinguish;
    // for each wgt k=1..n, why=(x*256+k)*8+i, i=0 rtn type, >0 rtn param i;
    //   i=7 after wgt is created, so you can modify it as needed
    return 0;} //~UserMacro

  // This event is enabled by Gtype&256:

  /**
  * Collided is called when a Bump-enabled (Collidable) sprite overlaps
  * another Bump-enabled widget, or if it runs off the edge of the game
  * board (if it is also Bump-enabled).
  *
  * @param  whom   reference to the sprite widget,
  * @param  hitt   reference to the widget it collided with,
  *                  or null of it ran off the board edge
  *
  * @return        true to stop looking (for more collisions on whom)
  */
  public boolean Collided(GameWgt whom, GameWgt hitt) { //     // (GameEv)
    return true;} //~Collided // return true to stop looking (for whom)

  // These are enabled by Gtype&512:

  /**
  * BlurEv is called for a widget when it loses focus. Your code may
  * also call BlurEv to unconditionally terminate focus for any other
  * reason. You should call super to do the default defocussing
  * (including telling the widget that it is losing focus) unless you
  * want to cancel the blur. If you cancel a blur that is called before
  * another widget is about to gain focus, then that focus is also
  * cancelled.
  *
  * @param  whom  The widget losing focus, or null if unknown by caller
  */
  public void BlurEv(GameWgt whom) { //                        // (GameEv)
    int refn = 0, why = 0, cx = 0;
    boolean rez;
    ClickLocn = -1;
    if (whom==null) whom = FocusWgt;
      else why = 2;
    if (whom != null) {
      refn = whom.GetRefNum();
      rez = whom.SetFocus(false,0,0);
      why = why+4;} //~if // why = +4
    if (whom==FocusWgt) FocusWgt = null;
      else why++; // why = +1
    if (FocusWgt != null) {
      rez = FocusWgt.SetFocus(false,0,0);
      why = why+4;} //~if
    why = -why;
    if (Logging) System.out.println("  (GmEv.BlrEv #" + refn + ") = "
        + why + "/" + cx);
    DrugWgtOfx = -1;
    FocusWgt = null;} //~BlurEv

  /**
  * FocusEv is called for any widget that is "Focusable" when the user
  * clicks on it. Your code may also call FocusEv to give focus to a
  * widget for any other reason. You must call super to do the default
  * focus processing unless you choose to cancel it. The base class
  * calls the whom.SetFocus method for this widget, and the widget also
  * has the opportunity to cancel (refuse) the focus.
  *
  * Only one widget at a time can have focus, so BlurEv is called for
  * any previous widget with focus before any other widget can get it.
  * If BlurEv (or the widgent it calls) refuses to give up focus,
  * then the new widget does not receive focus.
  *
  * @param  whom  The widget to be given focus
  * @param  vert  The vertical offset into this widget where
  *                 it was clicked
  * @param  horz  The horizontal offset into this widget
  */
  public void FocusEv(GameWgt whom, int vert, int horz) { //   // (GameEv)
    int refn = 0, why = 0;
    while (true) { // (vert,horz are wgt-relative)
      why++; // why = 1
      if (whom==null) { // why = 1, Bloo=1..
        if (FocusWgt != null) BlurEv(null);
        break;} //~if
      why++; // why = 2
      refn = whom.GetRefNum();
      if (FocusWgt != whom) {
        if (FocusWgt != null) BlurEv(null);
        if (FocusWgt != null) break;} //~if // blur was refused, can't refocus
      why++; // why = 3
      FocusWgt = whom;
      if (!whom.SetFocus(true,vert,horz)) {
        FocusWgt = null;
        break;} //~if
      why++; // why = 4
      if (ClickLocn<0) if (vert >= 0) {
        ClickLocn = (vert<<16)+horz;
        why++;} //~if // why = 5
      why = -why;
      break;} //~while
    if (Logging) System.out.println("  (GmEv.FocEv #" + refn + ") = "
        + why);} //~FocusEv

  /**
  * CancelDrag is called to cancel (terminate without effect) a drag. The
  * default CancelDrag replaces the WgtBeingDrug (if non-null and matches
  * the parameter) to its starting location as specified in DrugWgtBase.
  *
  * @param  whom  The widget being dragged
  */
  public void CancelDrag(GameWgt whom) {
    ClickLocn = -1;
    if (whom != null) if (WgtBeingDrug==whom) {
      if (Logging)
        System.out.println("** DragCanc #" + whom.GetRefNum() + " **");
      whom.SetPosn(DrugWgtBase>>16,SignExtend(DrugWgtBase));} //~if
    WgtBeingDrug = null;
    DrugWgtOfx = -1;
    Wgt2BeDrug = null;} //~CancelDrag

  /**
  * GotEscKey is called when the user types the EscapeKey,
  * if FocusWgt is not null, or else if DefaultFocus is not null,
  * and that widget accepts keystrokes. The base class uses GotEscKey
  * to cancel an outstanding drag or open dialog, and to terminate
  * any focus, so you should call super if this is not yours.
  *
  * @param  whom  The widget that has focus
  *
  * @return       true if taken (by alert or input dlog)
  */
  public boolean GotEscKey(GameWgt whom) { // from KeyData     // (GameEv)
    int nx = EvsDlogOp, rx = nx, zx = 0x436E-0x10000; // "Cn"
    boolean doit = false;
    GameWgt nuly = null;
    if (rx !=0) { // = EvsDlogOp
      if (rx<0) {
        if (Logging) OpCloEvDlog(0);
          else EvsDlogOp = 0;
        rx = -rx;} //~if
      nuly = GetIxWgt(rx);} //~if
    if (whom==WgtBeingDrug) CancelDrag(whom);
    if (nuly != null) if (theGame != null) doit = true;
    if (Logging) {
      if (whom != null) rx = whom.GetRefNum();
      System.out.println("  (GmEv.EscKy #" + rx + ") " + doit + " " + nx);} //~if
    if (doit) if (nx>0) // nx<0: IsDlgEv already closed it
      zx = theGame. IsDlgEvent(nuly,-1,zx); // close it
    if (whom==FocusWgt) {
      BlurEv(whom);  } //~if
    return doit;} //~GotEscKey

  /**
  * GotEnterKey is called when the user types the EnterKey,
  * if FocusWgt is not null, or else if DefaultFocus is not null,
  * and that widget accepts keystrokes. If a standard dialog is open,
  * it is closed and returns true.
  *
  * @param  whom  The widget that has focus (ignored here)
  *
  * @return       true if taken (by alert or input dlog)
  */
  public boolean GotEnterKey(GameWgt whom) { // from KeyData   // (GameEv)
    int nx = EvsDlogOp, rx = nx, zx = 0x4F4B-0x10000; // "OK"
    boolean doit = false;
    GameWgt nuly = null;
    if (rx !=0) { // = EvsDlogOp
      if (rx<0) {
        if (Logging) OpCloEvDlog(0);
          else EvsDlogOp = 0;
        rx = -rx;} //~if
      nuly = GetIxWgt(rx);} //~if
    if (nuly != null) if (theGame != null) doit = true;
    if (Logging) {
      if (whom != null) rx = whom.GetRefNum();
      System.out.println("  (GmEv.EntKy #" + rx + ") " + doit
          + " " + nx );}
    if (doit) if (nx>0) // nx<0: IsDlgEv already closed it
      zx = theGame. IsDlgEvent(nuly,-1,zx); // close it
    return doit;} //~GotEnterKey

  /**
  * GotTabKey is called when the user types the TabKey,
  * if FocusWgt is not null, or else if DefaultFocus is not null,
  * and that widget accepts keystrokes.
  *
  * @param  whom  The widget that has focus
  */
  public void GotTabKey(GameWgt whom) {} //~GotTabKey // from KeyData

  /**
  * KeyFld is called for any widget that has focus when the user
  * types any other key. If FocusWgt is null and DefaultFocus is null,
  * or if that widget does not accept keystrokes, then KeyFld is not called.
  * The base class uses KeyFld to insert keystrokes in an entry text field,
  * so you should call super if this is not yours.
  *
  * @param  whom  The widget that has focus
  * @param  info  The key that was typed
  */
  public void KeyFld(GameWgt whom, char info) { // from KeyData   (GameEv)
    int logy = SeenLog&512, tye, here, thar, lxx, rx = 0, why = 0;
    boolean nuly = false, fixt = false;
    String aLine, msg = "(" + (int)info;
    GameWgt prnt = null;
    while (true) { // (vert,horz are wgt-relative) (no early return)
      why++; // why = 1
      if (whom==null) break;
      rx = whom.GetRefNum();
      why++; // why = 2
      tye = whom.GetType();
      if ((tye&255) !=4) break;
      why++; // why = 3
      prnt = GetIxWgt(whom.GetParent());
      if (prnt != null) if ((prnt.GetInfo()&0x3FFF) == 0x658)
        fixt = true; // fixed-length if (parent's) mID = `xxFx`
      if (MaxLogWgts<FustKeyLog) MaxLogWgts = FustKeyLog;
      FustKeyLog = 0;
      here = whom.GetSelim();
      thar = whom.GetInfo()&0x3FFF;
      aLine = whom.GetTextLn();
      lxx = StrLength(aLine);
      if (lxx==0) fixt = false;
      if (here<0) here = lxx;
      if (info >= ' ') {
        why++; // why = 4
        if (fixt) {
          why++; // why = 5
          if (lxx<2) {
            aLine = "";
            here = 0;} //~if
          else if (here >= lxx) {
            aLine = RestOf(1,aLine);
            if (here>0) here--;} //~if
          else aLine = Substring(0,lxx-1,aLine);
          lxx--;} //~if
        if (here >= lxx) aLine = aLine + info;
        else if (here==0) aLine = "" + info + aLine;
          else aLine = Substring(0,here,aLine) + info + RestOf(here,aLine);
        here++;
        nuly = true;} //~if
      else if (info == '\b') {
        why = 8;
        if (here>0) {
          if (here>1) aLine = Substring(0,here-1,aLine) + RestOf(here,aLine);
            else aLine = RestOf(here,aLine);
          if (fixt) aLine = " " + aLine;
            else here--;
          nuly = true;} //~if
        else break;} //~if
      else if (info == Arro_LEFT) {
        why = 7;
        if (here>0) here--;
          else break;} //~if
      else if (info == Arro_RIGHT) {
        why = 9;
        if (here<lxx) here++;
          else break;} //~if
      else break; // why = 3
      why = why+8;
      if (nuly) lxx = whom.PutTextLn(aLine);
      whom.SetBlinker(false,here);
      break;} //~while
    if (Logging||(logy==0)) { // logy=0 1st time only
      System.out.println("  (GmEv.KeyFld #" + rx + ") " + msg + ") = " + why);
      SeenLog = SeenLog|512;}} //~KeyFld

  // This event is enabled by Gtype&1024, frm MouseDisp:           // (GameEv)

  /**
  * ClickEvt is called for a widget that is "Clickable" when the user
  * clicks on it (mouse down followed by mouse up on the same widget).
  * FocusEv has already been called for this widget. The base class uses
  * ClickEvt to detect and process double-click and standard dialog
  * events, so you should call super if you want that to happen.
  *
  * You would normally override ClickEvt to process button clicks.
  *
  * The vertical and horizontal offsets are relative to the top-left
  * corner of this widget. Add whom.AddPrntsPos(whom.GetPosn()) to get
  * the position relative to the window.
  *
  * @param  whom  The widget that was clicked
  * @param  vert  The vertical offset into this widget where
  *                 it was clicked
  * @param  horz  The horizontal offset into this widget
  *
  * @return       true if taken (by 2-click or alert or input dlog)
  */
  public boolean ClickEvt(GameWgt whom, int vert, int horz) { //  (GameEv)
    boolean didit;
    int now = GetMilliSecs(), // MouseDisp already called FocusEv -> SetFocus
        prio = CluckTime, kx = 0, nx = 0, zx = 0,
        why = 0, logy = 0;
    while (true) {
      if ((SeenLog&1024)==0) logy = 12345;
      if (EvsDlogOp>0) if (theGame != null) {
        zx = theGame. IsDlgEvent(whom,vert,horz);
          // zx=1: OK, =2: Cancel, =0: still open, -1: not mine
        if (zx==1) {
          didit = GotEnterKey(whom);
          why--;} //~if
        else if (zx==2) {
          didit = GotEscKey(whom);
          why = -2;} //~if
        else why = -3; // clicks outside dlog ignored
        zx = 0;
        break;} //~if
      why++; // why = 1
      if (Clik2time==0) break; // =999ms (set Clik2time=0 to defeat 2-click)
      CluckTime = now;
      why++; // why = 2
      if (whom==null) break;
      nx = whom.GetRefNum();
      why++; // why = 3
      if (logy>0) if (logy != 12345) if ((logy>>16) != nx)
        if ((logy&0xFFFF) != nx) logy = 0;
      if (prio+Clik2time<now) break;
      why++; // why = 4
      if (whom != FocusWgt) break;
      why = 0;
      break;} //~while
    if (Logging || (logy>0)) {
      if (why==0) logy++;
      else if (why>0) if (why<3) if (logy != 12345) logy = 0;
      if (logy>12345) SeenLog = SeenLog|1024;
      if (FocusWgt != null) zx = FocusWgt.GetRefNum();
      if (DefaultFocus != null) kx = DefaultFocus.GetRefNum();
      if (logy>0)
        System.out.println(IffyStr(why==0," (clik2ent) #"," (clik2not) #")
            + nx + " fw=" + zx + " df=" + kx
            + " " + (prio+Clik2time-now) + " = " + why );}
    if (why !=0) return (why<0); // got 2-click in focus = enter key..
    didit = GotEnterKey(whom);
    CluckTime = 0;
    return true;} //~ClickEvt

  // These events are enabled by Gtype&2048, frm MouseDisp:

  /**
  * After a MouseDn is called for a widget that is "Dragable" and the
  * user moves the mouse at least eight pixels with the button still
  * pressed, then it becomes a drag and StartDrag is called. If you
  * override this method, you can either cancel the drag by returning
  * null (without calling super), or else substitute another (possibly
  * new) widget for (in the same location as) the clicked widget when you
  * call super, and the replacement will be what is dragged. If you do
  * this, there is some additional code needed in your MouseUp override
  * so that the Dropped event works correctly.
  *
  * The vertical and horizontal offsets are relative to the top-left
  * corner of this widget. Add whom.AddPrntsPos(whom.GetPosn()) to get
  * the position relative to the window.
  *
  * @param  whom  The widget that was clicked
  * @param  vert  The vertical offset into this widget where it
  *                 was clicked
  * @param  horz  The horizontal offset into this widget
  *
  * @return       The widget being dragged, or null if cancelled
  */
  public GameWgt StartDrag(GameWgt whom, int vert, int horz) { // (GameEv)
    // subclass is allowed to substitute another wgt for whom, then call super
    int logy = SeenLog&2048, rx = 0, whar = 0;
    WgtBeingDrug = whom; // o'ride must already have decided
    if (whom==null) return null; // o'ride must also set Lox/Hix if needed
    whar = whom.GetPosn();
    rx = whom.GetRefNum();
    if (Logging||(logy==0)) {
      SeenLog = SeenLog|2048;
      if (MaxLogWgts<8) if (MaxLogAnim>8) MaxLogWgts = MaxLogAnim;
      System.out.println("** DragBgn #" + rx + " = " + vert + "/" + horz
          + LogVH(" @ ",whar,LogVH(" => (",DrugWgtOfx,") **")));
      JavaGame. LogWgtTree(whom,1);} //~if // start indented
    return whom;} //~StartDrag // caller calls Draggin if whom not null

  /**
  * After a StartDrag has been called for a Dragable widget and the user
  * releases the mouse button, Dropped is called for the same widget.
  * If you override this method and do something with the dropped widget
  * that could invalidate further drags on it using the arrow keys, you
  * should also call BlurEv for it, or else make sure it cannot take
  * focus. The base class does nothing with Dropped, so it is not
  * necessary to call super.
  *
  * @param  whom  The widget that was dragged
  */
  public void Dropped(GameWgt whom) { //                       // (GameEv)
    // from MouseUp if was still dragging (now up to date), (o'ride to use)
    int logy = SeenLog&8192, whar = 0, rx = 0;
    if (whom==null) return;
    whar = whom.GetPosn();
    rx = whom.GetRefNum();
    if (Logging||(logy==0)) {
      SeenLog = SeenLog|8192;
      System.out.println("** Dropped #" + rx + LogVH(" @ ",whar,
          LogVH(" => (",DrugWgtOfx,") **")));
      JavaGame. LogWgtTree(whom,1);} //~if // starts indented
    } //~Dropped

  /**
  * MouseDn is called for a widget that is "Dragable" when the user
  * presses the (right) mouse button down on some widget. If the button
  * is released without moving the mouse, then it is just a ClickEvt
  * and no drag is attempted.
  *
  * The vertical and horizontal offsets are relative to the top-left
  * corner of this widget. Add whom.AddPrntsPos(whom.GetPosn()) to get
  * the position relative to the window.
  *
  * @param  whom  The widget that was clicked
  * @param  vert  The vertical offset into this widget where it
  *                 was clicked
  * @param  horz  The horizontal offset into this widget
  */
  public void MouseDn(GameWgt whom, int vert, int horz) {
    int whar = (vert<<16)+horz;
    if (whom==null) return;
    if (ClickLocn<0) if (whar >= 0) ClickLocn = whar;
    if ((whom.GetType()&2048)==0) return;
    // Wgt2BeDrug = whom; // (no, we use ClickWgt, set by MouseDisp)
    DrugWgtBase = whom.GetPosn(); // (original) parent-relative
    DrugWgtOfx = whar;} //~MouseDn

  /**
  * After a StartDrag has been called for a Dragable widget and the user
  * continues to move the mouse, Dragging is called for the same widget.
  * Override this method to do something with the dragged widget's new
  * position. Dragging in the base class updates the dragged widget's
  * position, so call super before you try to use that position.
  *
  * The vertical and horizontal offsets are relative to the top-left corner
  * of this widget before its position has been adjusted by super.
  *
  * To get the position relative to the window, you should calculate
  * whom.AddPrntsPos(whom.GetPosn()) before calling super.
  *
  * @param  whom  The widget that is being dragged
  * @param  vert  The vertical offset into this widget before it moves
  * @param  horz  The horizontal offset into this widget
  */
  public void Dragging(GameWgt whom, int vert, int horz) { //  // (GameEv)
    // when dragging a wgt, DrugWgtOfx is wgt-rel clikLoc @ msDn, and we try
    //   to move the wgt so each new wgt-rel V/H matches (unchanged) DrugWgtOfx
    int mn, mx, px, zx, nuly, whar, posn = (vert<<16)+horz,
        base = DrugWgtOfx, more = AddPair(posn,PairNeg(base)); // = moved
    if (whom==null) return;
    if (WgtBeingDrug != whom) return;
    if (more==0) return;
    whar = whom.GetPosn(); // this is what we adjust to keep V/H=base
    nuly = AddPair(whar,more); // this is where to send it to
    zx = -DrugWgtGrid; // constrain to grid (relative to 0/0=parent topleft)..
    if (DrugWgtGrid>1) if ((DrugWgtGrid&zx)==DrugWgtGrid) {
      mx = (DrugWgtGrid>>1)*0x10001; // round to nearest grid point..
      nuly = AddPair(nuly,mx)&((zx<<16)|zx&0xFFFF);} //~if
    if ((vDragLims|hDragLims) !=0) { // enforce constraints..
      px = nuly&0xFFFF;              // .. = min<<16 + max&FFFF
      if (vDragLims !=0) {
        zx = nuly&(-0x10000);
        mn = vDragLims&(-0x10000);
        mx = vDragLims<<16;
        if (zx>mx) nuly = mx|px;
        if (zx<mn) nuly = mn|px;} //~if
      else nuly = whar&(-0x10000) | px; // no change if lims=0
      px = nuly&(-0x10000);
      if (hDragLims !=0) {
        zx = SignExtend(nuly);
        mn = vDragLims>>16;
        mx = SignExtend(vDragLims);
        if (zx>mx) nuly = mx&0xFFFF | px;
        if (zx<mn) nuly = mn&0xFFFF | px;} //~if
      else nuly = whar&0xFFFF | px;} //~if
    px = nuly>>16;
    zx = SignExtend(nuly);
    if (Logging) System.out.println("** DragMo #" + whom.GetRefNum()
        + LogVH(" = ",posn,LogVH(" @ ",whar,LogVH(" (",base,
        LogVH(" + ",more,LogVH(") => ",nuly," **"))))) );
    whom.SetPosn(px,zx);} //~Dragging

  /**
  * MouseUp is called for a widget that is "Dragable" when the user
  * releases the mouse button over the same widget that was clicked or
  * dragged. If the mouse is moved away from the clicked widget (and
  * there is no drag in process) then MouseUp is not called (and there
  * is no consequent ClickEvt).
  *
  * The vertical and horizontal offsets are relative to the top-left
  * corner of this widget. Add whom.AddPrntsPos(whom.GetPosn()) to get
  * the position relative to the window.
  *
  * @param  whom  The widget that was clicked
  * @param  vert  The vertical offset into this widget where it
  *                 was clicked
  * @param  horz  The horizontal offset into this widget
  */
  public void MouseUp(GameWgt whom, int vert, int horz) { //   // (GameEv)
    int posn = (vert<<16)+horz;
    boolean seen = false;
    if (whom != null) if (WgtBeingDrug != null) {
      if (WgtBeingDrug==whom) {
        if (posn != DrugWgtOfx)
          Dragging(whom,vert,horz); // now up to date
        if (WgtBeingDrug==whom)
          Dropped(whom);
        if (Logging) System.out.println("** DragFin #" + whom.GetRefNum() + " **");
        seen = true;} //~if
      else if (ClickWgt==whom) { // shouldn't be happenning, but it does..
        if (theGame != null) { // convert v/h from ClkWg to WgBDrg..
          vert = JavaGame.  GetGameValue(8); // = MouseVert
          horz = JavaGame.  GetGameValue(9); // = MouseHorz
          horz = theGame.  LocalizePos(WgtBeingDrug,vert,horz);
          vert = JavaGame.  GetGameValue(7); // LocalVert
          MouseUp(WgtBeingDrug,vert,horz);
          return;} //~if
        else CancelDrag(WgtBeingDrug);} //~if
      else CancelDrag(WgtBeingDrug);} //~if
    ClickLocn = -1;
    WgtBeingDrug = null;
    Wgt2BeDrug = null;
    if (Logging) if (!seen) if (whom != null)
      System.out.println("** MousUp #" + whom.GetRefNum() + " **");} //~MouseUp

  // This also is enabled by Gtype&512:                            // (GameEv)

  /**
  * GotArrowKey is called for any widget that has focus when the user
  * types one of the arrow keys. If FocusWgt is null and DefaultFocus
  * is null, or if that widget does not accept keystrokes, then
  * GotArrowKey is not called. The base class uses GotArrowKey to move
  * a dragable widget one pixel at time, so you should call super if
  * this is not yours.
  *
  * @param  whom  The widget that has focus
  * @param  info  The arrow key that was pressed, one of the constant
  *               values JavaGame.Arro_UP, JavaGame.Arro_DOWN,
  *               JavaGame.Arro_LEFT, or JavaGame.Arro_RIGHT.
  */
  public void GotArrowKey(GameWgt whom, char info) { //        // (GameEv)
    int tye = 0, refn = 0, whar = 0, why = 0;
    GameWgt prio = null;
    if (whom==null) return;
    tye = whom.GetType();
    if ((tye&767)==516) if ((FocusWgt==whom)||(DefaultFocus==whom))
        if ((info == Arro_RIGHT)||(info == Arro_LEFT)) {
      KeyFld(whom,info);
      return;} //~if
    if ((tye&2048)==0) return;
    whar = DrugWgtOfx; // (still valid after mouseUp)
    if (whar<0) return;
    if (info == Arro_DOWN) whar = whar+0x10000;
    else if (info == Arro_UP) whar = whar-0x10000;
    else if (info == Arro_RIGHT) whar++;
    else if (info == Arro_LEFT) whar--;
      else return;
    refn = whom.GetRefNum();
    prio = WgtBeingDrug;
    WgtBeingDrug = whom;
    Dragging(whom,whar>>16,SignExtend(whar));
    if (Logging)
      System.out.println("** DragKey #" + refn + " ** = " + why);
    WgtBeingDrug = prio;} //~GotArrowKey

  /**
  * KeyData is called to dispatch keystrokes if FocusWgt is not null,
  * or else if DefaultFocus is not null, and that widget accepts
  * keystrokes. KeyData in the base class separates out the control key
  * to call their specified event methods, so if you override KeyData, you
  * should call super if you want that to happen and this is not yours.
  *
  * @param  whom  The widget that has focus
  * @param  info  The key that was pressed
  */
  public void KeyData(GameWgt whom, char info) { //            // (GameEv)
    int tye = 0, refn = 0, whar = 0, why = 0;
    GameWgt prio = null; // if (WgtBeingDrug==whom) // (NOT)
    boolean seen = false;
    while (true) {
      why++; // why = 1
      if (whom != null) refn = whom.GetRefNum();
      if ((info == '\r')||(info == '\n')) {
        seen = GotEnterKey(whom);
        seen = false;
        break;} //~if
      why++; // why = 2
      if (info == '\t') {
        GotTabKey(whom);
        break;} //~if
      why++; // why = 3
      if (info == Esc_Key) {
        seen = GotEscKey(whom);
        seen = false;
        break;} //~if
      why++; // why = 4
      if (whom==null) break;
      why++; // why = 5
      tye = whom.GetType();
      // if (info == Esc_Key) why = 27;
      if (info == Arro_LEFT) why = 28;
      else if (info == Arro_RIGHT) why = 29;
      else if (info == Arro_UP) why = 30;
      else if (info == Arro_DOWN) why = 31;
      if (why>27) if (why<32) {
        GotArrowKey(whom,info);
        break;} //~if // why = 5
      why++; // why = 6
      if ((FocusWgt==whom)||(DefaultFocus==whom))
        KeyFld(whom,info);
      else why++; // why = 7
      break;} //~while    // seen = false;..
    if (Logging) if (!seen) System.out.println("** KyData #" + refn + " ** = "
        + (int)info
        + " = " + why);} //~KeyData

  // These are enabled by Gtype&4096: [untested]

  /**
  * RollIn is called when the user first rolls the mouse over a widget
  * that accepts rollover events.
  *
  * @param  whom  The widget that has focus and the mouse is over it
  * @param  vert  The vertical offset of the mouse in this widget
  * @param  horz  The horizontal offset of the mouse in this widget
  */
  public void RollIn(GameWgt whom, int vert, int horz) {} //~RollIn

  /**
  * RollOver is called each subsequent frame that the mouse is over
  * a widget that was sent a RollIn event.
  *
  * @param  whom  The widget that has focus and the mouse is over it
  * @param  vert  The vertical offset of the mouse in this widget
  * @param  horz  The horizontal offset of the mouse in this widget
  */
  public void RollOver(GameWgt whom, int vert, int horz) {} //~RollOver

  /**
  * RollOut is called the first frame that the mouse is no longer over
  * a widget that was sent a RollIn event.
  *
  * @param  whom  The widget that has focus and the mouse is not over it
  * @param  vert  The vertical offset of the mouse relative to this
  *                 widget but now outside its rectangle
  * @param  horz  The horizontal offset of the mouse relative to this
  *                 widget but outside its rectangle
  */
  public void RollOut(GameWgt whom, int vert, int horz) {} //~RollOut

/**
  * Default constructor.
  */
  public GameEvent() {
    ClickLocn = -1;
    DrugWgtOfx = -1;
    DrugWgtBase = 0;
    EvsDlogOp = 0;
    CluckTime = 0;
    vDragLims = 0;
    hDragLims = 0;
    DrugWgtGrid = 0;
    theGame = null;
    ClickWgt = null;
    RollyWgt = null;
    FocusWgt = null;
    Wgt2BeDrug = null;
    WgtBeingDrug = null;
    DefaultFocus = null;
    // MyDrugWgt = null;
    System.out.println("(new GamEvt)");}} //~GameEvent

// Index to Selected Useful GameEvent Methods (cuz BlueJ has none)..
//   OpCloEvDlog  GetRefWgt  GameList
//   Collided  GotEscKey  KeyFld  ClickEvt
//   Dropped  Dragging  GotArrowKey  KeyData

  // (javagame) (GE)
